#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# A widget to change date and time.
# 
# Author: JM. Philippe <jean-michel.philippe@doudoulinux.org>
#################################
# This file is part of DoudouLinux.
#
# DoudouLinux is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DoudouLinux is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with DoudouLinux.  If not, see <http://www.gnu.org/licenses/>.

############################
import sys, os, re, subprocess, gtk, gettext, datetime, pytz, gobject
from common import DataDir, set_gettext_domain

############################
############################
############################
class DatetimeWidget(object):
    """
    Date and time settings widget, to be used in a dedicated tool to set
    the keyboard layout, in the live installer, in the live relocalization
    tool, in the first boot of a computer, etc.
    
    The interface is contained in the root VBox returned by KbdWidget.root().
    """
    
    ############################
    def __init__(self, Window, helptext="tooltip", country=None):
        """
        Creates a date and time settings widget. The UTC help text can
        be embedded in a tooltip or in a label in the frame if there is
        enough space in the GUI.
        
        helptext = "tooltip"  the UTC help text is a tooltip
        helptext = "label"    the UTC help text is a label in the frame
        """
        # initializations
        self.timezone = ""
        self.timezone_code = ""
        self.tzdelta = datetime.timedelta(0)
        self.helptext = helptext
        
        # set the current time
        curtime = datetime.datetime.now()
        self.hours = curtime.hour
        self.minutes = curtime.minute
        self.seconds = curtime.second
        
        # guess language if necessary
        if country is None:
            self.country = os.environ["LANG"].split(".")[0]
            self.country = self.country.split("_")[1]
        else:
            self.country = country
        
        # find timezone
        fileid = open("/etc/timezone")
        self.timezone = fileid.read().strip()
        fileid.close()
        
        # create the interface
        self.window = Window
        self.gui = {}
        self.build_interface()
        self.init_interface()
        self.update_clock_delta()
        self.increment_clock()
        
    ############################
    def build_interface(self):
        """
        Creates the interface widgets.
        """
        # create widgets
        self.gui['/'] = gtk.VBox()
        self.gui['/time/'] = gtk.Frame()
        self.gui['/time/hbox/'] = gtk.HBox()
        self.gui['/time/hbox/icon'] = gtk.Image()
        self.gui['/time/hbox/box/'] = gtk.VBox()
        self.gui['/time/hbox/box/time/'] = gtk.HBox()
        self.gui['/time/hbox/box/time/hours'] = gtk.SpinButton()
        self.gui['/time/hbox/box/time/labhrs'] = gtk.Label()
        self.gui['/time/hbox/box/time/minutes'] = gtk.SpinButton()
        self.gui['/time/hbox/box/time/labmnts'] = gtk.Label()
        self.gui['/time/hbox/box/time/seconds'] = gtk.Label()
        self.gui['/time/hbox/box/time/labsecds'] = gtk.Label()
        self.gui['/time/hbox/box/check'] = gtk.CheckButton()
        self.gui['/time/hbox/box/help'] = gtk.Label()
        self.gui['/calendar'] = gtk.Calendar()
        
        # place the boxes
        self.gui['/'].pack_start(self.gui['/time/'], True, True, 2)
        self.gui['/'].pack_start(self.gui['/calendar'], True, True, 2)
        
        # place the time widgets
        self.gui['/time/'].add(self.gui['/time/hbox/'])
        self.gui['/time/hbox/'].pack_start(self.gui['/time/hbox/icon'], True, True, 0)
        self.gui['/time/hbox/'].pack_start(self.gui['/time/hbox/box/'], True, True, 10)
        box = self.gui['/time/hbox/box/time/']
        if self.helptext == "label":
            extraspace = 0
        else:
            extraspace = 5
        self.gui['/time/hbox/box/'].pack_start(box, True, True, extraspace)
        box.pack_start(self.gui['/time/hbox/box/time/hours'], False, False, 0)
        box.pack_start(self.gui['/time/hbox/box/time/labhrs'], False, False, 2)
        box.pack_start(self.gui['/time/hbox/box/time/minutes'], False, False, 0)
        box.pack_start(self.gui['/time/hbox/box/time/labmnts'], False, False, 2)
        box.pack_start(self.gui['/time/hbox/box/time/seconds'], False, False, 0)
        self.gui['/time/hbox/box/'].pack_start(self.gui['/time/hbox/box/check'], False, False, extraspace)
        self.gui['/time/hbox/box/'].pack_start(self.gui['/time/hbox/box/help'], True, True, 5)

    ############################
    def init_interface(self):
        """
        Initializes the graphical interface.
        """
        set_gettext_domain()
        
        # time
        self.gui['/time/'].set_label( _("Current time") )
        self.gui['/time/hbox/icon'].set_from_file( os.path.join(DataDir, "icons", "clock.png") )
        self.gui['/time/hbox/box/time/labhrs'].set_label(":")
        self.gui['/time/hbox/box/time/labmnts'].set_label(":")
        self.gui['/time/hbox/box/time/seconds'].set_label("00")
        self.gui['/time/hbox/box/time/hours'].set_range(0, 23)
        self.gui['/time/hbox/box/time/minutes'].set_range(0, 59)
        for button in (self.gui['/time/hbox/box/time/hours'], self.gui['/time/hbox/box/time/minutes']):
            button.set_digits(0)
            button.set_numeric(True)
            button.set_value(10)
            button.set_increments(1, 10)
            button.set_width_chars(2)
            button.connect('output', self.show_leading_zeros)
        self.gui['/time/hbox/box/time/hours'].connect("value-changed", self.set_hours)
        self.gui['/time/hbox/box/time/minutes'].connect("value-changed", self.set_minutes)
        
        # UTC
        self.gui['/time/hbox/box/check'].set_label( _("Use UTC time for the computer clock") )
        self.gui['/time/hbox/box/check'].connect("clicked", self.update_clock_delta)
        Text = _(u"Computers have an internal clock powered with batteries in order to retrieve the time at boot. This clock can use the universal time UTC or the local time that depends on your timezone. Usually Linux systems are using UTC while Windows® is using local time. If the time is not correct, try changing the UTC setting.").encode("utf-8")
        if self.helptext == "label":
            self.set_wrapped_label(self.gui['/time/hbox/box/help'], Text)
        else:
            self.gui['/time/hbox/box/check'].set_tooltip_text(Text)
        
        # set UTC check button, set time
        if os.path.isfile("/etc/default/rcS"):
            if not subprocess.call("grep -qs UTC=yes /etc/default/rcS", shell=True):
                self.gui['/time/hbox/box/check'].set_active(True)
        self.update_clock_delta()
        self.set_clock()

    ############################
    def root(self):
        """
        Returns the root widget of the keyboard settings widget. This is a Gtk VBox.
        """
        return self.gui['/']

    ############################
    def set_wrapped_label(self, widget, text):
        """
        Recomputes the label width whenever the window has been resized.
        """
        widget.set_label(text)
        widget.set_line_wrap(True)
        widget.set_size_request( self.window.get_size()[0] - 85, -1)

    ############################
    def show_leading_zeros(self, button):
        """
        Forces to display a leading zero.
        """
        adjustment = button.get_adjustment()
        button.set_text(
            "%02d" % int(adjustment.get_value())
        )
        return True

    ############################
    def update_clock_delta(self, *args):
        # if system is set with hwclk = local time
        # then if hwclk was UTC, we would read UTC
        # so we have to make from the current time a time object that is said to be UTC
        # if system is set with hwclk = UTC, then
        # then if hwclk was UTC, we would read the local time
        # so the hwclk can be deduced by subtracting the TZ delta
        
        # check if liveCD is set with hwclk = UTC or local time
        fileid = open("/etc/default/rcS")
        text = fileid.read()
        fileid.close()
        isUTC = re.findall('UTC=(.+)', text)
        if len(isUTC) and isUTC[0].strip() == "yes":
            isUTC = True
        else:
            isUTC = False
        
        # convert current time into an UTC time object
        ltime = list( datetime.datetime.now().timetuple() )
        ltime.pop(-1)
        ltime[-1] = pytz.timezone('UTC')
        ctime = datetime.datetime(*ltime)
        
        # compute the TZ delta by setting timezone then forcing UTC
        # to the result to be able to perform a datetime diff
        if self.timezone is not None:
            tz = pytz.timezone(self.timezone)
            ltime = list( ctime.astimezone(tz).timetuple() )
            ltime.pop(-1)
            ltime[-1] = pytz.timezone('UTC')
            tzdelta = datetime.datetime(*ltime) - ctime
        else:
            tzdelta = datetime.timedelta(0)
        
        # compute displayed time delta
        self.utcclock = self.gui['/time/hbox/box/check'].get_active()
        if not isUTC and self.utcclock:
            #ctime = ctime + tzdelta
            self.tzdelta = tzdelta
        elif isUTC and not self.utcclock:
            #ctime = ctime - tzdelta
            self.tzdelta = -tzdelta
        else:
            self.tzdelta = datetime.timedelta(0)
        self.set_clock()

    ############################
    def delta_hhmm(self):
        """
        Computes the hours and minutes of the current timezone delta.
        """
        # negative deltas lead to -1 day
        deltah = self.tzdelta.seconds + 24 * 3600 * self.tzdelta.days
        return deltah / 3600, deltah % 3600

    ############################
    def set_clock(self, hours=None, minutes=None, seconds=None):
        """
        Changes the displayed time.
        """
        # change HHMMSS
        if hours is not None:
            self.hours = hours
        if minutes is not None:
            self.minutes = minutes
        if seconds is not None:
            self.seconds = seconds
        
        # disconnect value-changed signals
        try:
            self.gui['/time/hbox/box/time/hours'].disconnect_by_func(self.set_hours)
            reconnectHours = True
        except:
            reconnectHours = False
        try:
            self.gui['/time/hbox/box/time/minutes'].disconnect_by_func(self.set_minutes)
            reconnectMin = True
        except:
            reconnectMin = False
        
        # update time
        deltah = self.delta_hhmm()
        self.gui['/time/hbox/box/time/hours'].set_value(self.hours + deltah[0])
        self.gui['/time/hbox/box/time/minutes'].set_value(self.minutes + deltah[1])
        self.gui['/time/hbox/box/time/seconds'].set_label("%02d" % self.seconds)
        
        # reconnect value-changed signals
        if reconnectHours:
            self.gui['/time/hbox/box/time/hours'].connect("value-changed", self.set_hours)
        if reconnectMin:
            self.gui['/time/hbox/box/time/minutes'].connect("value-changed", self.set_minutes)

    ############################
    def increment_clock(self):
        """
        Increments the displayed time of one second.
        Calls itself after one second is elapsed.
        """
        if self.seconds < 59:
            self.seconds += 1
        else:
            self.seconds = 0
            if self.minutes < 59:
                self.minutes += 1
            else:
                self.minutes = 0
                if self.hours < 23:
                    self.hours += 1
                else:
                    self.hours = 0
        self.set_clock()
        gobject.timeout_add_seconds(1, self.increment_clock)

    ############################
    def set_hours(self, widget):
        """
        Called whenever the hours are changed.
        """
        # we need to subtract the timedelta to the displayed time
        deltah = self.delta_hhmm()
        self.set_clock( hours=widget.get_value() - deltah[0])

    ############################
    def set_minutes(self, widget):
        """
        Called whenever the minutes are changed.
        """
        # we need to subtract the timedelta to the displayed time
        deltah = self.delta_hhmm()
        self.set_clock( minutes=widget.get_value() - deltah[1])

    ############################
    def set_datetime(self):
        """
        Changes the date and time.
        """
        if os.environ["USER"] != "root":
            print "ERROR: only root can set the system wide timezone."
            return
        
        year, month, day = self.gui['/calendar'].get_date()
        deltah = self.delta_hhmm()
        hour = self.hours + deltah[0]
        minutes = self.minutes + deltah[1]
        seconds = self.seconds
        
        command = "date %02d%02d%02d%02d%d.%02d" % (month+1, day, hour, minutes, year, seconds)
        print command
        subprocess.call(command, shell=True)
