#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# A widget to change keyboard layout.
# 
# Authors:
#   Clement Lefebvre <root@linuxmint.com>
#   JM. Philippe <jean-michel.philippe@doudoulinux.org>
#################################
# This file is part of DoudouLinux.
#
# DoudouLinux is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DoudouLinux is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with DoudouLinux.  If not, see <http://www.gnu.org/licenses/>.

############################
import sys, os, re, subprocess, gtk, gettext, common
from xml.dom.minidom import parse
from keyboard_layout import Keyboard

############################
# the system wide keyboard configuration file
KBDCONF = "/etc/default/keyboard"

############################
############################
############################
class KbdWidget(object):
    """
    Keyboard settings widget, to be used in a dedicated tool to set
    the keyboard layout, in the live installer, in the live relocalization
    tool, in the first boot of a computer, etc.
    
    The interface is contained in the root VBox returned by KbdWidget.root().
    """
    ############################
    def __init__(self, Window):
        """
        Creates a keyboard settings widget.
        """
        # initializations
        self.is_root = os.environ["USER"] == "root"
        self.initial_keyboard_layout = None
        self.initial_keyboard_model = None
        self.initial_keyboard_variant = None
        self.keyboard_layout = ""
        self.keyboard_layout_description = ""
        self.keyboard_model = ""
        self.keyboard_model_description = ""
        self.keyboard_variant = ""
        self.keyboard_variant_description = ""
        self.altkbd_layout = ""
        self.altkbd_variant = ""
        
        # create the interface
        self.window = Window
        self.gui = {}
        self.build_interface()
        self.init_interface()
        
    ############################
    def set_keyboard(self, layout=None, alt_layout="", variant=None, alt_variant="",model=None):
        """
        Changes keyboard settings for the current user.
        """
        # create the command line
        Args = []
        Command = "setxkbmap"
        if layout is not None:
            if alt_layout:
                Args.append(layout + alt_layout)
            else:
                Args.append(layout + self.altkbd_layout)
            Command += " -layout %s"
            print "Selected layout(s):", Args[-1]
        if variant is not None:
            if alt_variant:
                Args.append(variant + alt_variant)
            else:
                Args.append(variant + self.altkbd_variant)
            Command += " -variant %s"
            print "Selected variant(s):", Args[-1]
        if model is not None:
            Args.append(model)
            Command += " -model %s"
        
        # execute the command
        if Args:
            #print Command % tuple(Args)
            subprocess.call(Command % tuple(Args), shell=True)

    ############################
    def set_system_wide(self):
        """
        Sets the system wide keyboard settings.
        """
        if not self.is_root:
            return
        
        global KBDCONF
        Command = "sed -ri 's/^(%s)=.*$/\\1=\"%s\"/' " + KBDCONF
        Settings = (
            ("XKBLAYOUT", self.keyboard_layout + self.altkbd_layout),
            ("XKBVARIANT", self.keyboard_variant + self.altkbd_variant),
            ("XKBMODEL", self.keyboard_model)
        )
        for Key, Value in Settings:
            if Value is not None:
                print Command % (Key, Value)
                subprocess.call(Command % (Key, Value), shell=True)
        
    ############################
    def build_interface(self):
        """
        Creates the interface widgets.
        """
        # create widgets
        self.gui['/'] = gtk.VBox()
        self.gui['/lists/'] = gtk.HBox()
        self.gui['/lists/layouts/'] = gtk.ScrolledWindow()
        self.gui['/lists/layouts/tree'] = gtk.TreeView()
        self.gui['/lists/variants/'] = gtk.ScrolledWindow()
        self.gui['/lists/variants/tree'] = gtk.TreeView()
        self.gui['/layout/'] = gtk.VBox()
        self.gui['/layout/image'] = gtk.Frame()
        self.gui['/test/'] = gtk.HBox()
        self.gui['/test/label'] = gtk.Label()
        self.gui['/test/entry'] = gtk.Entry()
        self.gui['/model/'] = gtk.HBox()
        self.gui['/model/label'] = gtk.Label()
        self.gui['/model/combobox'] = gtk.ComboBox()
        
        # place the boxes
        self.gui['/'].pack_start(self.gui['/lists/'], True, True, 4)
        self.gui['/'].pack_start(self.gui['/layout/'], False, False, 2)
        self.gui['/'].pack_start(self.gui['/test/'], False, False, 2)
        self.gui['/'].pack_start(self.gui['/model/'], False, False, 2)
        
        # place the lists
        self.gui['/lists/'].pack_start(self.gui['/lists/layouts/'], True, True, 0)
        self.gui['/lists/layouts/'].add(self.gui['/lists/layouts/tree'])
        self.gui['/lists/'].pack_start(self.gui['/lists/variants/'], True, True, 0)
        self.gui['/lists/variants/'].add(self.gui['/lists/variants/tree'])
        self.gui['/lists/variants/tree'].set_size_request(350, -1)
        
        # place the layout image
        self.gui['/layout/'].pack_start(self.gui['/layout/image'], True, True, 0)
        
        # place the test box
        self.gui['/test/'].pack_end(self.gui['/test/entry'], True, True, 0)
        self.gui['/test/'].pack_end(self.gui['/test/label'], False, False, 4)
        
        # place the model combobox
        self.gui['/model/'].pack_end(self.gui['/model/combobox'], False, False, 0)
        self.gui['/model/'].pack_end(self.gui['/model/label'], False, False, 4)

    ############################
    def init_interface(self):
        """
        Initializes the graphical interface.
        """
        # kb models
        cell = gtk.CellRendererText()
        self.gui['/model/combobox'].pack_start(cell, True)
        self.gui['/model/combobox'].add_attribute(cell, 'text', 0)        
        self.gui['/model/combobox'].connect("changed", self.assign_keyboard_model)

        # kb layouts
        ren = gtk.CellRendererText()
        self.column10 = gtk.TreeViewColumn(_("Country"), ren)
        self.column10.add_attribute(ren, "text", 0)
        self.gui['/lists/layouts/tree'].append_column(self.column10)
        self.gui['/lists/layouts/tree'].connect("cursor-changed", self.assign_keyboard_layout)
        
        ren = gtk.CellRendererText()
        self.column11 = gtk.TreeViewColumn(_("Keyboard variant"), ren)
        self.column11.add_attribute(ren, "text", 0)
        self.gui['/lists/variants/tree'].append_column(self.column11)
        self.gui['/lists/variants/tree'].connect("cursor-changed", self.assign_keyboard_variant)
        
        self.kbdWidget = Keyboard( self.gui['/layout/image'] )
        self.gui['/layout/image'].add(self.kbdWidget)
        self.kbdWidget.set_size_request(
            self.window.get_size()[0] - 30,
            0.85*self.window.get_size()[0] * 4/15 + 6
        )
        self.build_kb_lists()
        
        # update texts
        self.gui['/lists/'].set_tooltip_text(
            _("First select your country, then select your keyboard variant. You will see the corresponding layout in the keyboard representation below.")
        )
        self.gui['/layout/image'].set_tooltip_markup(
            _("This shows the selected keyboard layout. In white and green the standard upper and lower case letters, in blue the characters printed when the <i>AltGr</i> key is simultaneously pressed.")
        )
        self.gui['/test/label'].set_label( _("Test your keyboard layout") )
        self.gui['/model/label'].set_label( _("Keyboard model") )
        self.gui['/model/'].set_tooltip_text(
            _("You may want to choose a particular keyboard model. If you cannot find your keyboard, use a generic keyboard model.")
        )

    ############################
    def root(self):
        """
        Returns the root widget of the keyboard settings widget. This is a Gtk VBox.
        """
        return self.gui['/']

    ############################
    def get_current_layout(self):
        """
        Gets the current keyboard settings and records them at the first run.
        """
        # analyse the current keyboard settings
        p = subprocess.Popen("setxkbmap -print", shell=True, stdout=subprocess.PIPE)
        keyboard_geom = "pc105"
        for line in p.stdout:
            # search for layout and variant
            if "xkb_symbols" in line:
                section = line.split("\"")[1]
                layout = section.split("+")[1]
                self.keyboard_layout = layout.split("(")[0]
                if "(" in layout:
                    self.keyboard_variant = layout.split("(")[1][:-1]
                else:
                    self.keyboard_variant = ""
                
                # search for alternative layout(s)
                settings = re.findall("\+([^:+]+):\d", line)
                if not self.altkbd_layout and settings:
                    for layout in settings:
                        self.altkbd_layout += ","
                        self.altkbd_layout += layout.split("(")[0]
                        self.altkbd_variant += ","
                        if "(" in layout:
                            self.altkbd_variant += layout.split("(")[1][:-1]
                    print self.altkbd_layout, self.altkbd_variant
            
            # search for keyboard model
            if "xkb_geometry" in line:
                first_bracket = line.index("(") +1
                substr = line[first_bracket:]
                last_bracket = substr.index(")")
                substr = substr[0:last_bracket]
                keyboard_geom = substr
        p.poll()
        
        # record the initial settings
        if self.initial_keyboard_layout is None:
            self.initial_keyboard_layout = self.keyboard_layout
        if self.initial_keyboard_variant is None:
            self.initial_keyboard_variant = self.keyboard_variant
        if self.initial_keyboard_model is None:
            self.initial_keyboard_model = keyboard_geom
        
        return keyboard_geom

    ############################
    def build_kb_lists(self):
        """
        Builds the keyboard layouts list.
        """

        # firstly we'll determine the layouts in use
        keyboard_geom = self.get_current_layout()

        xml_file = '/usr/share/X11/xkb/rules/xorg.xml'
        model_models = gtk.ListStore(str,str)
        model_models.set_sort_column_id(0, gtk.SORT_ASCENDING)
        model_layouts = gtk.ListStore(str,str)
        model_layouts.set_sort_column_id(0, gtk.SORT_ASCENDING)        
        dom = parse(xml_file)

        # if we find the users keyboard info we can set it in the list
        set_keyboard_model = None
        set_keyboard_layout = None
        set_keyboard_variant = None

        # grab the root element
        root = dom.getElementsByTagName('xkbConfigRegistry')[0]
        # build the list of models
        root_models = root.getElementsByTagName('modelList')[0]
        for element in root_models.getElementsByTagName('model'):
            conf = element.getElementsByTagName('configItem')[0]
            name = conf.getElementsByTagName('name')[0]
            desc = conf.getElementsByTagName('description')[0]
            #vendor = conf.getElementsByTagName('vendor')[0] # presently unused..
            iter_model = model_models.append([self.get_xml_text(desc.childNodes), self.get_xml_text(name.childNodes)])
            item = self.get_xml_text(name.childNodes)
            if item == keyboard_geom:
                set_keyboard_model = iter_model
        root_layouts = root.getElementsByTagName('layoutList')[0]
        for element in root_layouts.getElementsByTagName('layout'):
            conf = element.getElementsByTagName('configItem')[0]
            name = conf.getElementsByTagName('name')[0]
            desc = conf.getElementsByTagName('description')[0]
            iter_layout = model_layouts.append([self.get_xml_text(desc.childNodes), self.get_xml_text(name.childNodes)])
            item = self.get_xml_text(name.childNodes)
            if item == self.keyboard_layout:
                set_keyboard_layout = iter_layout
        # now set the model        
        self.gui['/model/combobox'].set_model(model_models)
        self.gui['/lists/layouts/tree'].set_model(model_layouts)

        if set_keyboard_layout is not None:
            # show it in the list
            treeview = self.gui['/lists/layouts/tree']
            model = treeview.get_model()
            column = treeview.get_column(0)
            path = model.get_path(set_keyboard_layout)
            treeview.set_cursor(path, focus_column=column)
            treeview.scroll_to_cell(path, column=column)
        if set_keyboard_model is not None:         
             # show it in the combo
            combo = self.gui['/model/combobox']
            model = combo.get_model()                    
            combo.set_active_iter(set_keyboard_model)            
            
    ############################
    def build_kb_variant_lists(self):
        """
        Builds the keyboard variants list.
        """
        # firstly we'll determine the layouts in use
        keyboard_geom = self.get_current_layout()

        xml_file = '/usr/share/X11/xkb/rules/xorg.xml'      
        model_variants = gtk.ListStore(str,str)
        model_variants.set_sort_column_id(0, gtk.SORT_ASCENDING)        
        dom = parse(xml_file)
        
        # grab the root element
        root = dom.getElementsByTagName('xkbConfigRegistry')[0]
        # build the list of variants       
        root_layouts = root.getElementsByTagName('layoutList')[0]
        for layout in root_layouts.getElementsByTagName('layout'):
            conf = layout.getElementsByTagName('configItem')[0]
            layout_name = self.get_xml_text(conf.getElementsByTagName('name')[0].childNodes)            
            layout_description = self.get_xml_text(conf.getElementsByTagName('description')[0].childNodes)            
            if layout_name != self.keyboard_layout:
                continue
            
            iter_variant = model_variants.append([layout_description, None])  
            variants_list = layout.getElementsByTagName('variantList')
            if len(variants_list) <= 0:
                continue
            
            root_variants = layout.getElementsByTagName('variantList')[0]   
            for variant in root_variants.getElementsByTagName('variant'):                    
                variant_conf = variant.getElementsByTagName('configItem')[0]
                variant_name = self.get_xml_text(variant_conf.getElementsByTagName('name')[0].childNodes)
                variant_description = "%s - %s (%s)" % (
                    layout_description,
                    self.get_xml_text(variant_conf.getElementsByTagName('description')[0].childNodes),
                    variant_name
                )
                iter_variant = model_variants.append([variant_description, variant_name])                                                    
            break
                                                                                
        # now set the model        
        self.gui['/lists/variants/tree'].set_model(model_variants)
        
        # select the first item (standard variant layout)
        treeview = self.gui['/lists/variants/tree']
        model = treeview.get_model()
        column = treeview.get_column(0)
        path = model.get_path(model.get_iter_first())
        treeview.set_cursor(path, focus_column=column)
       
    ############################
    def get_xml_text(self, nodelist):
        """
        Extracts text from Xml text nodes.
        """
        rc = []
        for node in nodelist:
            if node.nodeType == node.TEXT_NODE:
                rc.append(node.data)
        return ''.join(rc)
    
    ############################
    def assign_keyboard_model(self, combobox, data=None):
        """
        Called whenever someone updates the keyboard model.
        """
        model = combobox.get_model()
        active = combobox.get_active()
        if active > -1:
            row = model[active]
            self.set_keyboard(model=row[1])
            self.keyboard_model = row[1]
            self.keyboard_model_description = row[0]

    ############################
    def assign_keyboard_layout(self, treeview, data=None):
        """
        Called whenever someone updates the keyboard layout.
        """
        model = treeview.get_model()
        active = treeview.get_selection().get_selected_rows()
        if len(active) < 1:
            return
        active = active[1][0]
        if active is None:
            return
        row = model[active]
        self.set_keyboard(layout=row[1])
        self.keyboard_layout = row[1]
        self.keyboard_layout_description = row[0]
        self.build_kb_variant_lists()

    ############################
    def assign_keyboard_variant(self, treeview, data=None):
        """
        Called whenever someone updates the keyboard variant.
        """
        model = treeview.get_model()
        active = treeview.get_selection().get_selected_rows()
        if len(active) < 1:
            return
        active = active[1][0]
        if(active is None):
            return
        row = model[active]
        if row[1] is None:
            self.set_keyboard(layout=self.keyboard_layout)
            self.keyboard_variant = ""
        else:
            self.set_keyboard(variant=row[1])
            self.keyboard_variant = row[1]
        self.keyboard_variant_description = row[0]
        self.kbdWidget.setLayout(self.keyboard_layout)
        self.kbdWidget.setVariant(self.keyboard_variant)

    ############################
    def set_size_request(self, Width, Height):
        """
        Changes the keyboard widget size request.
        You have to provide Width, Height that fit your application window.
        """
        self.kbdWidget.set_size_request(Width, Height)

    ############################
    def set_initial(self):
        """
        Sets the current selection to the initial keyboard settings.
        """
        # set the initial layout
        treeview = self.gui['/lists/layouts/tree']
        model = treeview.get_model()
        iter = model.get_iter_first()                
        while iter is not None:
            iter_layout = model.get_value(iter, 1)
            if iter_layout.lower() == self.initial_keyboard_layout.lower():
                column = treeview.get_column(0)
                path = model.get_path(iter)
                treeview.set_cursor(path, focus_column=column)
                treeview.scroll_to_cell(path, column=column)
                break
            iter = model.iter_next(iter)
        
        # set the initial variant
        if self.initial_keyboard_variant is None:
            return
        treeview = self.gui['/lists/variants/tree']
        model = treeview.get_model()
        iter = model.get_iter_first()                
        while iter is not None:
            iter_variant = model.get_value(iter, 1)
            if iter_variant is not None and iter_variant.lower() == self.initial_keyboard_variant.lower():
                column = treeview.get_column(0)
                path = model.get_path(iter)
                treeview.set_cursor(path, focus_column=column)
                treeview.scroll_to_cell(path, column=column)
                break
            
            iter = model.iter_next(iter)

    ############################
    def restore(self):
        """
        Restores the initial keyboard settings.
        """
        if self.initial_keyboard_variant:
            self.set_keyboard(
                model=self.initial_keyboard_model,
                layout=self.initial_keyboard_layout,
                variant=self.initial_keyboard_variant
            )
        else:
            self.set_keyboard(
                model=self.initial_keyboard_model,
                layout=self.initial_keyboard_layout
            )
