#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# Authors:
#   Clement Lefebvre <root@linuxmint.com>
#   JM. Philippe <jean-michel.philippe@doudoulinux.org>
#################################
# This file is part of DoudouLinux.
#
# DoudouLinux is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DoudouLinux is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with DoudouLinux.  If not, see <http://www.gnu.org/licenses/>.

############################
import subprocess
import sys, os, types, gtk, cairo
from gtk import gdk
from math import pi

############################
############################
############################
def fromUnicodeString(raw):
    """
    Converts an Unicode value "U+…" or "+U+…" into a character.
    """
    if raw[0:2] == "U+":
        return unichr(int(raw[2:], 16))
    elif raw[0:2] == "+U":
        return unichr(int(raw[3:], 16))
    elif raw == "dead_circumflex":
        return '^'
    elif raw == "dead_diaeresis":
        return '¨'
    elif raw == "dead_tilde":
        return '~'
    else:
        return ""


############################
############################
############################
class CairoWidget(gtk.DrawingArea):
	"""
	A Cairo base widget on which to draw. Draws nothing interesting.
	"""
	
	# draw in response to an expose-event
	__gsignals__ = { "expose-event": "override" }
	
	############################
	def __init__(self):
		"""
		A base widget with signals ready to be overriden.
		"""
		super(CairoWidget, self).__init__()
		
		# gtk.Widget signals
		self.connect("button_press_event", self.button_press)
		self.connect("button_release_event", self.button_release)
		self.connect("motion_notify_event", self.motion_notify)
		
		# notification of parent application
		self.notify_callback = None
		self.notify_event = "unset"
		
		# More GTK voodoo : unmask events
		self.add_events(gdk.BUTTON_PRESS_MASK |
						gdk.BUTTON_RELEASE_MASK |
						gdk.POINTER_MOTION_MASK)
	
	############################
	def do_expose_event(self, event):
		"""
		Handle the expose-event by drawing.
		"""
		# Create the cairo context
		self.cr = self.window.cairo_create()
		
		# Restrict Cairo to the exposed area; avoid extra work
		#self.cr.rectangle(event.area.x, event.area.y, event.area.width, event.area.height)
		#self.cr.clip()
		
		self.draw()
	
	############################
	def draw(self):
		"""
		Fill the background with gray.
		"""
		width, height = self.window.get_size()
		self.cr.set_source_rgb(0.5, 0.5, 0.5)
		self.cr.rectangle(0, 0, width, height)
		self.cr.fill()
	
	############################
	def notify(self, callback, eventName):
		"""
		Records a callback to notify the parent application of changes.
		"""
		self.notify_callback = callback
		self.notify_event = eventName
	
	############################
	def button_press(self, widget, event):
		pass
	def button_release(self, widget, event):
		pass
	def motion_notify(self, widget, event):
		pass


############################
############################
############################
class Keyboard(CairoWidget):
    """
    A graphical keyboard layout widget.
    """

    # keycodes of keyboard models
    kb_104 = {
        "extended_return": False,
        "keys": [
        (0x29, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa, 0xb, 0xc, 0xd),
        (0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x2b),
        (0x1e, 0x1f, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28),
        (0x2c, 0x2d, 0x2e, 0x2f, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35),
        ()]
    }

    kb_105 = {
        "extended_return": True,
        "keys": [
        (0x29, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa, 0xb, 0xc, 0xd),
        (0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b),
        (0x1e, 0x1f, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x2b),
        (0x54, 0x2c, 0x2d, 0x2e, 0x2f, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35),
        ()]
    }

    kb_106 = {
        "extended_return": True,
        "keys": [
        (0x29, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe),
        (0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b),
        (0x1e, 0x1f, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29),
        (0x2c, 0x2d, 0x2e, 0x2f, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36),
        ()]
    }
    
    # we need to change of font depending on the language alphabet
    # since the default font can't display all alphabets
    font_family_table = {
        "default": "DejaVu Sans",
        "ben": "MuktiNarrow",
        "ben_probhat": "MuktiNarrow",
        "bolnagri": "gargi",
        "guj": "Rekha",
        "hin-wx": "gargi",
        "hy": "Artsounk",
        "in": "gargi",
        "kan": "Kedage",
        "mal": "Meera",
        "mal_lalitha": "Meera",
        "mm": "Ayar Bon Bon",
        "tam": "Lohit Tamil",
        "tam_keyboard_with_numerals": "Lohit Tamil",
        "tam_TAB": "TAMu_Kadambri",
        "tam_TSCII": "TSCu_Times",
        "tam_unicode": "Lohit Tamil",
        "tel": "Pothana2000",
        "th": "Loma",
    }

    ############################
    def __init__(self, container, layout="us", variant=""):
        """
        Creates a keyboard layout widget in an existing container,
        with given keyboard layout and variant.
        """
        super(Keyboard, self).__init__()
        
        # default dimensions
        self.margin = 6
        self.width = 500
        self.height = 150
        
        # initializations
        self.codes = []
        self.keypaths = {}      # Cairo paths of keys without glyphs
        self.font_family = ""
        self.kb = None
        self.container = container
        self.layout = layout
        self.variant = variant
        
        # setup keyboard widget
        self.set_font_family()
        self.loadCodes()
        self.set_kbdmodel()
        self.do_size_request()
    
    ############################
    def set_font_family(self):
        """
        Sets the key font family using the global font table.
        """
        # try to match the layout then the variant
        if self.variant in self.font_family_table:
            self.font_family = self.font_family_table[self.variant]
        elif self.layout in self.font_family_table:
            self.font_family = self.font_family_table[self.layout]
        else:
            self.font_family = self.font_family_table["default"]

    ############################
    def setLayout(self, layout):
        """
        Sets the keyboard layout.
        """
        #print layout
        self.layout = layout
        self.set_font_family()

    ############################
    def setVariant(self, variant):
        """
        Sets the keyboard variant.
        """
        #print variant
        if variant is None:
            self.variant = ""
        else:
            self.variant = variant
        self.loadCodes()
        self.set_kbdmodel()
        self.set_font_family()
        self.queue_draw()

    ############################
    def set_kbdmodel(self):
        """
        Sets the keyboard model and eventually clears the Cairo keys.
        """
        kbl_104 = ["us", "th"]
        kbl_106 = ["jp"]
        prev_kb = self.kb

        # most keyboards are 105 key so default to that
        if self.layout in kbl_104:
            self.kb = self.kb_104
        elif self.layout in kbl_106:
            self.kb = self.kb_106
        elif self.kb != self.kb_105:
            self.kb = self.kb_105
        
        # number of keys changed
        if self.kb != prev_kb:
            self.keypaths = {}
    
    ############################
    def set_size_request(self, w, h):
        """
        Resizes the keyboard widget by resizing its parent container
        and redrawing all the keys.
        """
        self.width = int(w)
        self.height = int(h)
        self.keypaths = {}  # we need to redraw the keys totally
        self.container.set_size_request(self.width, self.height)
        self.do_size_request()
    
    ############################
    def do_size_request(self, *args):	
        """
        Recomputes widget dimensions.
        """
        # recompute key sizes
        self.key_w = (self.width - 16 * self.margin) / 15
        kh = (self.height - 5 * self.margin) / 4
        if kh < self.key_w:
            self.key_w = kh
        
        # recompute x offset to center keyboard in its parent
        self.x0 = (self.width - 15 * (self.key_w + self.margin)) / 2
        
        self.queue_draw()

    ############################
    def roundedRect(self, x, y, width, height, radius):
        """
        Creates a Cairo rounded rectangle.
        """
        # the Cairo path of the rectangle is stored to reduce computation effort
        pathname = "+%d+%d %dx%d r%d" % (x, y, width, height, radius)
        
        if pathname in self.keypaths:
            # try to use an already computed Cairo path
            self.cr.append_path( self.keypaths[pathname] )
        else:
            # compute the path
            self.cr.arc(x + radius, y + radius, radius, 2*(pi/2), 3*(pi/2))
            self.cr.arc(x + width - radius, y + radius, radius, 3*(pi/2), 4*(pi/2))
            self.cr.arc(x + width - radius, y + height - radius, radius, 0*(pi/2), 1*(pi/2))
            self.cr.arc(x + radius, y + height - radius, radius, 1*(pi/2), 2*(pi/2))
            self.cr.close_path()
            self.keypaths[pathname] = self.cr.copy_path_flat()
        
        # set path color
        self.cr.set_source_rgb(0.35, 0.35, 0.35)
        self.cr.fill()
    
    ############################
    def draw_text(self, text, x, y, xalign, yalign):
        """
        Draws the text of a given key.
        """
        # get glyphs dimensions
        fascent, fdescent, fheight, fxadvance, fyadvance = self.cr.font_extents()
        xbearing, ybearing, width, height, xadvance, yadvance = self.cr.text_extents(text)
        
        # compute x offset
        space = 0.1*self.key_w
        if xalign == "left":
            #dx = fxadvance/2 - width/2
            dx = fxadvance / 2 - xadvance / 2 - space
        elif xalign == "right":
            #dx = self.key_w - fxadvance/2 - width/2
            dx = self.key_w - xadvance + xbearing - space
        else:
            print "ERROR: no xalign value", xalign
            return
        
        # compute y offset
        space = 0.05*self.key_w
        if yalign == "top":
            #dy = fheight - fdescent
            dy = fascent + space
        elif yalign == "bottom":
            #dy = self.key_w - fdescent
            dy = self.key_w - fdescent - space
        else:
            print "ERROR: no yalign value", yalign
            return
        
        # draw text in key
        self.cr.move_to(x + dx, y + dy)
        self.cr.show_text(text)
        self.cr.stroke()
    
    ############################
    def drawRow(self, keylist, x, y, rx, last_end=False):
        """
        Draws a row of the keyboard.
        """
        rw = 15 * self.key_w + 14 * self.margin
        for key in keylist:
            self.roundedRect(x, y, self.key_w, self.key_w, rx)

            self.cr.select_font_face(self.font_family, cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_BOLD)
            self.cr.set_source_rgb(1, 1, 1)
            self.draw_text(self.regular_text(key), x, y, "left", "bottom")

            self.cr.set_source_rgb(0.76, 1, 0.16)
            self.draw_text(self.shift_text(key), x, y, "left", "top")

            self.cr.select_font_face(self.font_family, cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_NORMAL)
            #self.cr.set_source_rgb(0x49, 0x7f, 0x79))
            self.cr.set_source_rgb(0.61, 0.7, 0.87)
            self.draw_text(self.alt_text(key), x, y, "right", "bottom")

            #self.cr.set_source_rgb(0.41, 0.5, 0.67)
            self.draw_text(self.shift_alt_text(key), x, y, "right", "top")

            rw = rw - self.margin - self.key_w
            x = x + self.margin + self.key_w

        return (x,rw)
    
    ############################
    def draw(self):
        """
        Draws the whole keyboard.
        """
        # compute key size
        width, height = self.container.get_size_request()
        kw = self.key_w
        
        # set font size
        self.font_size = (kw - 2*self.margin) / 2
        self.cr.set_font_size(self.font_size)
        
        # clear the keyboard area
        self.cr.set_source_rgb(0xd6, 0xd6, 0xd6)
        self.cr.rectangle(0, 0, width, height)
        self.cr.fill()
        
        # initializations
        keys = self.kb["keys"]
        ext_return = self.kb["extended_return"]
        x = self.x0
        y = self.margin
        rx = 0.175 * self.key_w
        first_key_w = 0
        rows = 4
        
        # draw rows of keys
        remaining_x = [0,0,0,0]
        remaining_widths = [0,0,0,0]
        for nrow in range(0, rows):
            # compute width of first key, on the very left
            if first_key_w > 0:
                first_key_w = first_key_w * 1.375
                if self.kb == self.kb_105 and nrow == 3:
                    first_key_w = kw * 1.275

                self.roundedRect(x, y, first_key_w, kw, rx)
                x += first_key_w + self.margin
            else:
                first_key_w = kw
            
            # draw the current row
            x, rw = self.drawRow(keys[nrow], x, y, rx, nrow==1 and not ext_return)
            
            # compute space left, on the right
            remaining_x[nrow] = x
            remaining_widths[nrow] = rw
            if nrow:
                remaining_widths[nrow] -= first_key_w + self.margin
            
            # the missing key is not the return key, draw it
            if nrow != 1 and nrow != 2:
                self.roundedRect(x, y, remaining_widths[nrow], kw, rx)
            
            # prepare next loop
            x = self.x0
            y = y + self.margin + kw
        
        # we now have to draw the return key
        if ext_return:
            # compute corners location
            x1 = remaining_x[1]
            y1 = self.margin + kw*1 + self.margin*1
            w1 = remaining_widths[1]
            x2 = remaining_x[2]
            y2 = self.margin + kw*2 + self.margin*2
            
            # this is some serious crap... but it has to be so
            # maybe one day keyboards won't look like this...
            # one can only hope
            if "return" in self.keypaths:
                # try to reuse a previously computed Cairo path
                self.cr.append_path( self.keypaths["return"] )
            else:
                # compute the Cairo path of the return key
                self.cr.arc(x1+rx,    y1+rx,    rx, 2*(pi/2), 3*(pi/2))
                self.cr.arc(x1+w1-rx, y1+rx,    rx, 3*(pi/2), 4*(pi/2))
                self.cr.arc(x1+w1-rx, y2+kw-rx, rx, 0*(pi/2), 1*(pi/2))
                self.cr.arc(x2+rx,    y2+kw-rx, rx, 1*(pi/2), 2*(pi/2))
                self.cr.arc_negative(x2-rx, y1+kw+rx, rx, 4*(pi/2), 3*(pi/2))
                self.cr.arc(x1+rx,    y1+kw-rx, rx, 1*(pi/2), 2*(pi/2))
                self.cr.close_path()
                self.keypaths["return"] = self.cr.copy_path_flat()
            
            # draw the path
            self.cr.fill()
        else:
            # much simpler key to draw!
            x = remaining_x[2]
            y = self.margin + (kw + self.margin)*2
            self.roundedRect(x, y, remaining_widths[2], kw, rx)

    ############################
    def snapshot(self, filename, filetype='png'):
        """
        Takes a PNG snapshot of the keyboard.
        """
        if not filename:
            return
        
        # compute snapshoting area
        x = self.x0 - self.margin
        y = 0
        w = 15 * self.key_w + 16 * self.margin
        h = 4 * self.key_w + 5 * self.margin
        
        # perform snapshot
        pixbuf = gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, False, 8, w, h)
        pixmap = self.get_snapshot( gdk.Rectangle(x, y, w, h) )
        pixbuf.get_from_drawable(pixmap, self.get_colormap(), 0, 0, 0, 0, w, h)
        pixbuf.save(filename, filetype)

    ############################
    def regular_text(self, index):
        """
        Returns the lowercase character corresponding to a given key.
        """
        return self.codes[index - 1][0]

    ############################
    def shift_text(self, index):
        """
        Returns the uppercase character corresponding to a given key.
        """
        return self.codes[index - 1][1]

    ############################
    def ctrl_text(self, index):
        """
        Returns the character corresponding to a given key + Ctrl.
        """
        return self.codes[index - 1][4]

    ############################
    def alt_text(self, index):
        """
        Returns the character corresponding to a given key + AltGr.
        """
        return self.codes[index - 1][2]

    ############################
    def shift_alt_text(self, index):
        """
        Returns the character corresponding to a given key + Shift + AltGr.
        """
        return self.codes[index - 1][3]

    ############################
    def loadCodes(self):
        """
        Loads the codes of the current keyboard layout/variant.
        """
        if self.layout is None:
            return
        
        # get key codes
        variantParam = ""
        if self.variant:
            variantParam = "-variant %s" % self.variant
        cmd="ckbcomp -model pc106 -layout %s %s -compact" % (self.layout, variantParam)
        #print cmd
        pipe = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=None)
        cfile = pipe.communicate()[0]

        #clear the current codes
        del self.codes[:]
        
        # fill the key characters table
        for l in cfile.split('\n'):
            if l[:7] != "keycode":
                continue
            
            # extract characters
            codes = l.split('=')[1].strip().split(' ')
            plain = fromUnicodeString(codes[0])
            shift = fromUnicodeString(codes[1])
            ctrl = fromUnicodeString(codes[2])
            alt = fromUnicodeString(codes[3])
            
            # append characters to code list
            if ctrl == plain:
                ctrl = ""
            if alt == plain:
                alt = ""
            self.codes.append((plain, shift, ctrl, alt))


############################
############################
############################
if __name__ == "__main__":
    def quit(*args):
        gtk.main_quit()
    
    # create a window
    window = gtk.Window()
    window.set_size_request(600, 200)
    window.connect("delete-event", quit)
    window.set_title( "Keyboard layout" )
    
    vbox = gtk.VBox()
    container = gtk.Frame()
    #container.set_size_request(600, 300)
    window.add(container)

    layout=sys.argv[1]
    variant = sys.argv[2]
    filename = sys.argv[3]

    kb1 = Keyboard(container, layout, variant)
    container.add(kb1)
    vbox.pack_start(container, True, True, 0)
    
    def resize(window, *args):
        w, h = window.get_size()
        kb1.set_size_request(w - 10, h - 10)
    window.connect('size-allocate', resize)
    
    window.show_all()
    window.present()

    kb1.snapshot(filename)
    gtk.main()
