#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# A widget to change keyboard layout.
# 
# Authors:
#   Clement Lefebvre <root@linuxmint.com>
#   JM. Philippe <jean-michel.philippe@doudoulinux.org>
#################################
# This file is part of DoudouLinux.
#
# DoudouLinux is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DoudouLinux is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with DoudouLinux.  If not, see <http://www.gnu.org/licenses/>.

############################
import sys, os, re, subprocess, gtk, gettext, urllib2, glob
from common import DataDir, set_gettext_domain, geoloc_country

############################
LOCALEGEN = "/etc/locale.gen"
LOCALEDEF = "/etc/default/locale"
FLAGDIR = os.path.join(DataDir, "flags")
COUNTRYLIST = os.path.join(DataDir, 'countries')
LANGLIST = os.path.join(DataDir, 'languages')
DoudouLinuxL10n = "/usr/lib/doudoulinux-l10n/update-l10n"

############################
# use region flags instead of country flag for few languages
REGIONFLAGS = {
    "ast": "_Asturias",
    "ca":  "_Catalonia",
    "cy":  "_Wales",
    "eu":  "_Basque Country",
    "gl":  "_Galicia",
    "gd":  "_Scotland",
}

############################
# some locales use an alternative alphabet whose name needs to be translated
gettext.install("iso_15924")  # switch to alphabet name translations
VARIANTLIST = {
    "cyrillic": "Cyrillic",
    "devanagari": "Devanagari (Nagari)",
    "latin": "Latin",
    "iqtelif": "İQTElif",
}
# this is a trick to avoid xgettext extract messages
# that are picked from the iso-codes MO files
for key, value in VARIANTLIST.items():
    VARIANTLIST[key] = _(value)

############################
############################
############################
class LangWidget(object):
    """
    Language settings widget, to be used in a dedicated tool to set
    the system language, in the live installer, in the live relocalization
    tool, in the first boot of a computer, etc.
    
    The interface is contained in the root VBox returned by LangWidget.root().
    """
        
    ############################
    def __init__(self, Window, localefilt=None, all_locales=False, keep_locales=True):
        """
        Creates a language settings widget.
        
        localefilt = list of locales to show, eg. ['fr', 'en'], 'fr' will match 'fr_*'
        """
        # initializations
        self.localefilt = localefilt
        self.all_locales = all_locales
        self.keep_locales = keep_locales
        
        # create the interface
        self.window = Window
        self.gui = {}
        self.build_interface()
        self.init_interface()
        
    ############################
    def build_interface(self):
        """
        Creates the interface widgets.
        """
        # create widgets
        self.gui['/'] = gtk.VBox()
        self.gui['/label'] = gtk.Label()
        self.gui['/lang/'] = gtk.ScrolledWindow()
        self.gui['/lang/tree'] = gtk.TreeView()
        
        # place the boxes
        self.gui['/'].pack_start(self.gui['/label'], False, False, 4)
        self.gui['/'].pack_start(self.gui['/lang/'], True, True, 4)
        self.gui['/lang/'].add(self.gui['/lang/tree'])

    ############################
    def init_interface(self):
        """
        Initializes the graphical interface.
        """
        self.set_wrapped_label(
            self.gui['/label'],
            _("Choose the default language of the system. Please note that translations are not all available in every language."),
        )
        
        ren = gtk.CellRendererPixbuf()
        column = gtk.TreeViewColumn( _("Flag"), ren)
        column.add_attribute(ren, "pixbuf", 2)
        self.gui['/lang/tree'].append_column(column)

        ren = gtk.CellRendererText()
        column = gtk.TreeViewColumn( _("Language"), ren)
        column.add_attribute(ren, "text", 0)
        self.gui['/lang/tree'].append_column(column)
        self.gui['/lang/tree'].connect("cursor-changed", self.assign_language)
        
        # build the language list
        self.build_lang_list()

    ############################
    def set_wrapped_label(self, widget, text):
        """
        Recomputes the label width whenever the window has been resized.
        """
        widget.set_label(text)
        widget.set_line_wrap(True)
        widget.set_size_request( self.window.get_size()[0] - 85, -1)

    ############################
    def set_language(self):
        """
        Sets the system wide language.
        """
        global LOCALEGEN, LOCALEDEF
        if os.environ["USER"] != "root":
            print "ERROR: only root can set the system wide locale."
            return
        
        newlocale = "%s.UTF-8" % self.language
        
        # first update the locale list
        if not self.keep_locales:
            subprocess.call("sed -i 's/^[^#]/# &/' %s" % LOCALEGEN, shell=True)
        subprocess.call("sed -ri 's/^#\s*(%s.+UTF-8$)/\\1/' %s" % (self.language, LOCALEGEN), shell=True)
        
        # recompile locales
        if self.keep_locales:
            subprocess.call("locale-gen --keep-existing", shell=True)
        else:
            subprocess.call("locale-gen", shell=True)
        subprocess.call("update-locale --no-checks LANG", shell=True)
        subprocess.call("update-locale LANG=%s" % newlocale, shell=True)
        
        # schedule a GDM conf reload
        subprocess.call("/etc/init.d/gdm reload", shell=True)
        
        # update DoudouLinux system translations
        if os.path.isfile(DoudouLinuxL10n):
            subprocess.call(DoudouLinuxL10n, shell=True)
        
    ############################
    def root(self):
        """
        Returns the root widget of the keyboard settings widget. This is a Gtk VBox.
        """
        return self.gui['/']

    ############################
    def get_countries(self):
        """
        Gets the country list.
        """
        global COUNTRYLIST
        
        countries = {}
        file = open(COUNTRYLIST, "r")
        gettext.install("iso_3166")  # switch to country name translations
        for line in file:
            line = line.strip()
            split = line.split("=")
            if len(split) == 2:
                countries[split[0]] = _(split[1])
        file.close()
        
        set_gettext_domain()  # restore app translations
        return countries

    ############################
    def get_languages(self):
        """
        Gets the language list.
        """
        global LANGLIST
        
        languages = {}
        file = open(LANGLIST, "r")
        gettext.install("iso_639")  # switch to language name translations
        for line in file:
            line = line.strip()
            split = line.split("=")
            if len(split) == 2:
                languages[split[0]] = _(split[1])
        file.close()
        
        set_gettext_domain()  # restore app translations
        return languages
    
    ############################
    def build_lang_list(self):
        """
        Builds the locale list tree.
        """
        global REGIONFLAGS, VARIANTLIST, LOCALEGEN
        
        # find the country code and the language code
        cur_country_code = geoloc_country()
        cur_lang = os.environ['LANG']
        if "." in cur_lang:
            cur_lang = cur_lang.split(".")[0]
        
        # read countries and languages
        countries = self.get_countries()
        languages = self.get_languages()
        
        # tree initialization
        model = gtk.ListStore(str,str,gtk.gdk.Pixbuf)
        model.set_sort_column_id(0, gtk.SORT_ASCENDING)
        
        # create tree by reading the locales list
        locales = open(LOCALEGEN, "r")
        set_index = None
        found = []  # there may be two occurences of the same locale for an unknown reason
        for line in locales:
            # extract the language code
            match = re.findall("^(?:#\s+)?([^ \.]+).+UTF-8\s*$", line)
            if not match:
                continue
            locale_code = match[0]
            split = locale_code.split("_")
            if len(split) != 2:
                continue
            if locale_code in found:
                continue
            found.append(locale_code)
            
            # if there is no translation files, skip the language
            if not self.all_locales:
                localedir = "/usr/share/locale/%s/LC_MESSAGES" % locale_code
                localedir2 = "/usr/share/locale/%s/LC_MESSAGES" % split[0]
                if not os.path.isdir(localedir) and not os.path.isdir(localedir2):
                    continue
                if not glob.glob("%s/*" % localedir) and not glob.glob("%s/*" % localedir2):
                    continue
            
            # filter locales if requested
            if self.localefilt is not None:
                ismatching = False
                for langfilt in self.localefilt:
                    if len(langfilt) > len(locale_code):
                        continue
                    if locale_code[:len(langfilt)] == langfilt:
                        ismatching = True
                        break
                if not ismatching:
                    continue
            
            # find the language name
            language_code = split[0]
            if language_code in languages:
                language = languages[language_code]
            else:
                language = language_code
            
            # find the country name + language variant if any
            country_code = split[1].lower()
            if "@" in country_code:
                variant = country_code.split("@")[1]
                if variant in VARIANTLIST:
                    variant = ", " + VARIANTLIST[variant]
                else:
                    variant = ", " + variant.capitalize()
                country_code = country_code.split("@")[0]
            else:
                variant = ""
            if country_code in countries:
                country = countries[country_code]
            else:
                country = country_code
            
            # find the flag name and path
            if language_code in REGIONFLAGS:
                flag_name = REGIONFLAGS[language_code]
            else:
                flag_name = country_code
            flag_path = os.path.join(FLAGDIR, flag_name + '.png')
            if not os.path.exists(flag_path):
                flag_path = os.path.join(FLAGDIR, 'generic.png')
            
            # compute the language label to be displayed
            language_label = "%s (%s)%s" % (language, country, variant)
            
            # create a tree element
            iter = model.append()
            model.set_value(iter, 0, language_label)
            model.set_value(iter, 1, locale_code)
            model.set_value(iter, 2, gtk.gdk.pixbuf_new_from_file(flag_path))
            
            # select our country based on geoloc if it matches
            if cur_country_code is not None and cur_country_code.lower() == country_code:
                if set_index is None:
                    set_index = iter                                
                else:
                    # If we find more than one language for a particular country, one of them being English, go for English by default.
                    if language_code == "en":
                        set_index = iter                 
                    # Guesswork... handy for countries which have their own language (fr_FR, de_DE, es_ES.. etc. )
                    elif country_code == language_code:
                        set_index = iter
            
            # select our country based on the LANG system variable
            if set_index is None and locale_code == cur_lang:
                set_index = iter
                #print "Set via locale: " + cur_lang
        print "Found %d usable locales" % len(found)
        
        # update tree
        treeview = self.gui['/lang/tree']
        treeview.set_model(model)
        if set_index is not None:
            column = treeview.get_column(0)
            path = model.get_path(set_index)
            treeview.set_cursor(path, focus_column=column)
            treeview.scroll_to_cell(path, column=column)
        treeview.set_search_column(0)

    ############################
    def assign_language(self, treeview, data=None):
        """
        Called whenever someone updates the language
        """
        model = treeview.get_model()
        active = treeview.get_selection().get_selected_rows()
        if len(active) < 1:
            return
        active = active[1][0]
        if active is None:
            return
        
        row = model[active]
        self.language = row[1]
        print "Selected language:", self.language
        #try:            
            #self.translation = gettext.translation('live-installer', LOCALEDIR, languages=[self.setup.language])
            #self.translation.install()
        #except Exception, detail:
            #print "No translation found, switching back to English"
            #self.translation = gettext.translation('live-installer', LOCALEDIR, languages=['en'])
            #self.translation.install()        
