#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# A widget to change timezone. Requires kde-l10n to translate timezone names.
# 
# Authors:
#   Clement Lefebvre <root@linuxmint.com>
#   JM. Philippe <jean-michel.philippe@doudoulinux.org>
#################################
# This file is part of DoudouLinux.
#
# DoudouLinux is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DoudouLinux is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with DoudouLinux.  If not, see <http://www.gnu.org/licenses/>.

############################
import sys, os, re, subprocess, gtk, gettext, math, tempfile
from PIL import Image
from common import DataDir, set_gettext_domain, geoloc_country

TZDIR = os.path.join(DataDir, "timezone")

############################
############################
############################
class Timezone:
    """
    Timezone definition, with latitude, longitude, name, position on map, etc.
    """
    
    # guess image size
    tmpimg = Image.open(os.path.join(TZDIR, "cc.png"))
    width = tmpimg.size[0]  # Width of the map
    height = tmpimg.size[1] # Height of the map
    del tmpimg
    
    ############################
    def __init__(self, name, country_code, coordinates):
        """
        Creates a named timezone with its pixel coordinates on map.
        """
        # initializations
        self.name = name
        self.country_code = country_code
        self.coordinates = coordinates
        
        # extract latitude/longitude
        latlongsplit = coordinates.find('-', 1)
        if latlongsplit == -1:
            latlongsplit = coordinates.find('+', 1)
        if latlongsplit != -1:
            self.latitude = coordinates[:latlongsplit]
            self.longitude = coordinates[latlongsplit:]
        else:
            self.latitude = coordinates
            self.longitude = '+0'
        
        # convert latitude/longitude to integers
        self.latitude = self.parse_position(self.latitude, 2)
        self.longitude = self.parse_position(self.longitude, 3)
        
        # compute position in image coordinates
        self.x, self.y = self.getPosition(self.latitude, self.longitude)            
    
    ############################
    def parse_position(self, position, wholedigits):
        """
        Converts the string of longitude/latitude into a floating number.
        """
        if position == '' or len(position) < 4 or wholedigits > 9:
            return 0.0
        
        floatnbr = position[:wholedigits + 1] + "." + position[wholedigits + 1:]
        return float(floatnbr)
        
    ############################
    def getPosition(self, la, lo):
        """
        Computes pixel coordinate of a latitude and longitude.
        Uses Miller map projection, but is also clipped (eg. no Antartica).
        """
        # need to add/sub magic numbers because the map doesn't actually go from -180...180, -90...90
        # thus the upper corner is not -180, -90 and we have to compensate
        # we need a better method of determining the actually range so we can better place citites (shtylman)
        xdeg_offset = -6
        # the 180 - 35) accounts for the fact that the map does not span the entire -90 to 90
        # the map does span the entire 360 though, just offset
        x = (self.width * (180.0 + lo) / 360.0) + (self.width * xdeg_offset/ 180.0)
        x = x % self.width

        #top and bottom clipping latitudes
        topLat = 81
        bottomLat = -59

        #percent of entire possible range
        topPer = topLat/180.0

        # get the y in rectangular coordinates
        y = 1.25 * math.log(math.tan(math.pi/4.0 + 0.4 * math.radians(la)))

        # calculate the map range (smaller than full range because the map is clipped on top and bottom
        fullRange = 4.6068250867599998
        # the amount of the full range devoted to the upper hemisphere
        topOffset = fullRange*topPer
        mapRange = abs(1.25 * math.log(math.tan(math.pi/4.0 + 0.4 * math.radians(bottomLat))) - topOffset)

        # Convert to a percentage of the map range
        y = abs(y - topOffset)
        y = y / mapRange

        # this then becomes the percentage of the height
        y = y * self.height

        return (int(x), int(y))        


############################
############################
############################
class TzWidget(object):
    """
    Timezone settings widget, to be used in a dedicated tool to set
    the keyboard layout, in the live installer, in the live relocalization
    tool, in the first boot of a computer, etc.
    
    The interface is contained in the root VBox returned by KbdWidget.root().
    """
    
    # the timezone colors used to retrieve the timezone delta
    timezone_colors = {
        "2a0100": "-11.0",
        "560000": "-10.0",
        "66ff00": "-9.5",
        "800000": "-9.0",
        "ac0001": "-8.0",
        "d70000": "-7.0",
        "ff0000": "-6.0",
        "66ff00": "-5.5",
        "ff2327": "-5.0",
        "bdff00": "-4.5",
        "ff5255": "-4.0",
        "0cff00": "-3.5",
        "ff807e": "-3.0",
        "ffaaa7": "-2.0",
        "fcd6d6": "-1.0",
        "2a1200": "0.0",
        "562200": "1.0",
        "803300": "2.0",
        "ab4401": "3.0",
        "0bff63": "3.5",
        "d65400": "4.0",
        "00cdff": "4.5",
        "fe6700": "5.0",
        "1965ff": "5.5",       
        "00ffcb": "5.75",
        "ff7d27": "6.0",
        "cd00ff": "6.5",
        "ff9751": "7.0",
        "ffb381": "8.0",
        "fecca9": "9.0",
        "ad0047": "9.5",
        "fce7d6": "10.0",
        "d40658": "10.5",
        "d2ab00": "11.0",
        "ff0069": "11.5",
        "ffcc06": "12.0",
        "fd2c80": "12.75",
        "ff589c": "13.0",
        
        #Add some timezones for cities which are located on borders (for which the color doesn't match the color of the rest of the timezone)
        "856d64": "0.0",   # Africa/Bissau, X: 237, Y: 164
        "dbbbb6": "0.0",   # Africa/Conakry, X: 240, Y: 168
        "ff9696": "-3.0",   # America/Argentina/Buenos_Aires, X: 170, Y: 237
        "e37456": "-4.0",   # America/Curacao, X: 154, Y: 164
        "ff3d3e": "-6.0",   # America/Indiana/Knox, X: 126, Y: 115
        "ff7172": "-4.0",   # America/Iqaluit, X: 155, Y: 69
        "b1bb5f": "-3.5",   # America/Miquelon, X: 173, Y: 104
        "ff2e2d": "-5.0",   # America/Port-au-Prince, X: 149, Y: 154
        "ff6968": "-4.0",   # America/Rio_Branco, X: 156, Y: 197
        "ff4645": "-5.0",   # America/Thunder_Bay, X: 122, Y: 102
        "ff8230": "4.0",   # Asia/Baku, X: 338, Y: 117
        "fcac6c": "8.0",   # Asia/Kuching, X: 432, Y: 180
        "ff8b3c": "7.0",   # Asia/Novokuznetsk, X: 396, Y: 91
        "ffbc94": "8.0",   # Asia/Singapore, X: 422, Y: 180
        "6cd471": "4.0",   # Asia/Yerevan, X: 330, Y: 117
        "e28b5b": "10.5",   # Australia/Lord_Howe, X: 508, Y: 232
        "8b5c3e": "2.0",   # Europe/Sofia, X: 297, Y: 113
        "ffa468": "12.75",   # Pacific/Chatham, X: 547, Y: 254
        "ff589c": "13.0",   # Pacific/Tongatapu, X: 548, Y: 215
    }
    
    # images
    pil_images = {
        "cc.png": Image.open(os.path.join(TZDIR, "cc.png")).convert('RGB'),
        "bg.png": Image.open(os.path.join(TZDIR, "bg.png")).convert('RGBA'),
        "dot.png": Image.open(os.path.join(TZDIR, "dot.png")).convert('RGBA'),
    }
    
    ############################
    def __init__(self, Window, country=None, debug=False):
        """
        Creates a timezone settings widget.
        """
        # initializations
        self.timezone = ""
        self.timezones = []
        self.window = Window
        self.default_size = self.window.get_size()
        self.overlays = {}
        self.xoffset = 0
        self.yoffset = 0
        
        # create a file name used to save PNG temporary images
        tmpfile = tempfile.mkstemp(suffix=".png")
        self.tmpfile = tmpfile[1]
        os.close(tmpfile[0])
        
        # guess language if necessary
        if country is None:
            self.country = os.environ["LANG"].split(".")[0]
            self.country = self.country.split("_")[1]
        else:
            self.country = country
        
        # create the interface
        self.window = Window
        self.gui = {}
        self.build_interface()
        self.init_interface(debug=debug)
        self.default_timezone(self.country)
    
    ############################
    def set_offsets(self, xoffset, yoffset):
        """
        Sets map offsets.
        """
        self.xoffset = xoffset
        self.yoffset = yoffset
    
    ############################
    def exit(self):
        """
        Free objects before leaving.
        """
     
    ############################
    def set_timezone(self):
        """
        Sets the system wide timezone.
        """
        if os.environ["USER"] != "root":
            print "ERROR: only root can set the system wide timezone."
            return
        
        # check the localtime file exists
        localtime = "/usr/share/zoneinfo/%s" % self.timezone
        if not os.path.isfile(localtime):
            print "ERROR: cannot find file '%s'" % localtime
            return
        
        # set the timezone
        subprocess.call("cp %s /etc/localtime" % localtime, shell=True)
        subprocess.call("echo \"%s\" | tee /etc/timezone" % self.timezone, shell=True)
       
    ############################
    def build_interface(self):
        """
        Creates the interface widgets.
        """
        # create widgets
        self.gui['/'] = gtk.VBox()
        self.gui['/map/'] = gtk.EventBox()
        self.gui['/map/image'] = gtk.Image()
        self.gui['/select/'] = gtk.HBox()
        self.gui['/select/label'] = gtk.Label()
        self.gui['/select/gmt'] = gtk.Label()
        self.gui['/select/combo'] = gtk.ComboBox()
        
        # place the boxes
        self.gui['/'].pack_start(self.gui['/map/'], True, True, 4)
        self.gui['/'].pack_start(self.gui['/select/'], False, False, 2)
        
        # place the map
        self.gui['/map/'].add(self.gui['/map/image'])
        #self.gui['/lists/variants/tree'].set_size_request(350, -1)
        
        # place the select box
        self.gui['/select/'].pack_end(self.gui['/select/combo'], False, False, 10)
        self.gui['/select/'].pack_end(self.gui['/select/gmt'], False, False, 0)
        self.gui['/select/'].pack_end(self.gui['/select/label'], False, False, 4)
    
    ############################
    def init_interface(self, debug=False):
        """
        Initializes the graphical interface.
        """
        global TZDIR
        set_gettext_domain()
        
        # timezone combo
        self.combo_timezones = self.gui['/select/combo']
        self.combo_timezones.connect('changed', self.timezone_combo_selected)
        
        cell = gtk.CellRendererText()
        self.combo_timezones.pack_start(cell, True)
        self.combo_timezones.add_attribute(cell, 'text', 0)
        
        # timezone map
        self.timezone_map = self.gui['/map/image']
        self.timezone_map.set_from_file(os.path.join(TZDIR, "bg.png"))
        self.gui['/map/'].connect("button-release-event", self.timezone_map_clicked)
        self.gui['/map/'].set_tooltip_text(
            _("Click on the map to select your location.")
        )
        self.build_timezones(debug=debug)
                
        # timezone selected
        self.gui['/select/label'].set_label( _("Selected timezone:") )
        self.gui['/select/gmt'].set_label("GMT+0000")
        self.gui['/select/gmt'].set_tooltip_text(
            _("The timezone delta (+0100, -0800, etc.) does not take into account daylight saving time. This delta may then be one hour shifted compared to the real localtime.")
        )
        self.gui['/select/combo'].set_tooltip_text(
            _("Select your timezone. They are arranged by continent and are using the names of towns.")
        )
        
    ############################
    def root(self):
        """
        Returns the root widget of the keyboard settings widget. This is a Gtk VBox.
        """
        return self.gui['/']
        
    ############################
    def build_timezones(self, debug=False):
        """
        Builds the list of towns for each timezone.
        """
        # initialization of debug variables
        if debug:
            notfound = []
            outside = []
            notused = self.timezone_colors.keys()
        
        # populate timezones
        gettext.install("timezones4")  # switch to timezone names translations
        model = gtk.ListStore(str, object)
        model.set_sort_column_id(0, gtk.SORT_ASCENDING)                   
        timezones = open("/usr/share/zoneinfo/zone.tab", "r")
        for line in timezones:
            # select lines with relevant information
            if line.strip().startswith("#"):
                continue
            content = line.strip().split("\t")
            if len(content) < 2:                    
                continue
            
            # create a new town
            country_code = content[0]
            coordinates = content[1]
            timezone = content[2]
            tz = Timezone(timezone, country_code, coordinates)
            self.timezones.append(tz)
            iter = model.append()
            model.set_value(iter, 0, _(timezone).replace("/", " – ").replace("_", " ") )
            model.set_value(iter, 1, tz)  
            
            # collect information for the debug mode
            if not debug:
                continue
            #print "Timezone: %s, X: %s, Y: %s" % (tz.name, tz.x, tz.y)
            rgb_im = self.pil_images["cc.png"]
            try:
                hexcolor = '%02x%02x%02x' % rgb_im.getpixel((tz.x, tz.y))
            except:
                outside.append( "  %s, X: %s, Y: %s" % (tz.name, tz.x, tz.y) )
                continue
            #print " Color: #%s" % (hexcolor)
            
            if hexcolor not in self.timezone_colors:
                notfound.append( (hexcolor, "%s, X: %s, Y: %s" % (tz.name, tz.x, tz.y)) )
            elif hexcolor in notused:
                notused.remove(hexcolor)
            
        # update the combobox list
        set_gettext_domain()  # restore app translations
        self.combo_timezones.set_model(model)
        
        # print information about colors/coords not matched in the debug mode
        if not debug:
            return
        print "Coords outside of image:"
        print "\n".join(outside)
        
        print "Color not found: %d towns" % len(notfound)
        notfound.sort(cmp=lambda x, y: 2*int(x[1] > y[1]) - 1)
        Colors = []
        for Pair in notfound:
            print "  color %s, %s" % Pair
            if Pair[0] not in Colors:
                Colors.append(Pair[0])
        
        print "Missing colors: %d colors" % len(Colors)
        print "Extra colors: %d colors" % len(notused)
        print "\n".join(notused)
                     
    ############################
    def timezone_combo_selected(self, combobox):
        """
        Called whenever the combobox is changed.
        """
        model = combobox.get_model()
        index = combobox.get_active()
        if index:
            timezone = model[index][1]
            self.timezone_select(timezone)            
    
    ############################
    def timezone_map_clicked(self, widget, event):
        """
        Called whenever the map is clicked.
        """
        # compute coordinates corrected with image offset in event box
        x = event.x
        y = event.y
        xoffset = ( self.window.get_size()[0] - self.timezones[0].width ) / 2
        x -= xoffset + self.xoffset
        yoffset = ( self.window.get_size()[1] - self.default_size[1] ) / 2
        y -= yoffset + self.yoffset
        print "Coords: %s[+%s] %s[+%s]" % (x, xoffset, y, yoffset)
        
        # don't update the timezone if the user clicks outside the image
        if x < 0 or y < 0 or x > self.timezones[0].width or y > self.timezones[0].height:
            return
        
        # search for the nearest town
        min_distance = 1000 # Looking for min, starting with a large number
        closest_timezone = None
        for timezone in self.timezones:
            distance = abs(x - timezone.x) + abs(y - timezone.y)
            if distance < min_distance:
                min_distance = distance
                closest_timezone = timezone
        
        # set the town timezone as the current timezone
        print "Closest timezone %s" % closest_timezone.name
        self.timezone_select(closest_timezone)
        
        # update the combobox
        model = self.combo_timezones.get_model()
        iter = model.get_iter_first()
        while iter is not None:            
            if closest_timezone.name == model.get_value(iter, 1).name:
                self.combo_timezones.set_active_iter(iter)
                break
            iter = model.iter_next(iter)
        
    ############################
    def get_overlaid(self, overlay_path):        
        """
        Get the overlaid world map for a given timezone overlay image.
        """
        # compute the overlaid map whenever needed
        if overlay_path not in self.overlays:
            overlay = Image.open(overlay_path)
            overlay = overlay.convert("RGBA")
            self.overlays[overlay_path] = self.pil_images["bg.png"].copy()
            self.overlays[overlay_path].paste(overlay, (0,0), overlay)
        
        # return a copy of the background image
        return self.overlays[overlay_path].copy()
    
    ############################
    def timezone_select(self, timezone):        
        """
        Changes the current timezone.
        """
        global TZDIR
        
        # find the color of the timezone under the mouse
        rgb_im = self.pil_images["cc.png"]
        hexcolor = '%02x%02x%02x' % rgb_im.getpixel((timezone.x, timezone.y))
        #print "Color: #%s" % (hexcolor)
        
        # convert the color into the overlay image path
        GMT = self.timezone_colors[hexcolor]
        overlay_path = os.path.join(TZDIR, "timezone_%s.png" % GMT)
        #print "Image: %s" % overlay_path
        
        # overlay the picture of the timezone on the map
        dot = self.pil_images["dot.png"]
        background = self.get_overlaid(overlay_path)
        background.paste(dot, (timezone.x-3, timezone.y-3), dot)
        background.save(self.tmpfile, "PNG")
        self.timezone_map.set_from_file(self.tmpfile)
        
        # change the GTM text
        GMT = float(GMT)
        Minutes = 60 * (GMT - int(GMT))
        self.gui['/select/gmt'].set_label("GMT%+03d%02d" % (GMT, Minutes))
        
        # Save the selection
        self.timezone = timezone.name
        print "Selected timezone:", self.timezone
        
    ############################
    def default_timezone(self, language):        
        """
        Sets the default timezone from the system settings, the country code
        or the locale. The country code can be guessed using geolocalization.
        """
        # try the system settings
        timezone = "Etc/UTC"
        if os.path.isfile("/etc/timezone"):
            fileid = open("/etc/timezone")
            tzname = fileid.read().strip()
            fileid.close()
        if tzname != "Etc/UTC" and tzname in [x.name for x in self.timezones]:
            timezone = tzname
        
        # guess the country code from IP address
        # we give it a higher priority than system settings
        country_code = geoloc_country()
        if country_code is not None:
            timezone = "Etc/UTC"
            country_code = country_code.upper()
            print "Using geolocalization"
        
        # get the country code from locale code
        if country_code is None:
            if "_" in language:
                country_code = language.split("_")[1]                    
            else:
                country_code = language.upper()
        
        # change the current selection
        combo = self.gui['/select/combo']
        model = combo.get_model()
        iter = model.get_iter_first()
        print "country_code =", country_code
        while iter is not None:
            if timezone == "Etc/UTC":
                iter_country_code = model.get_value(iter, 1).country_code
                if iter_country_code == country_code:
                    combo.set_active_iter(iter)
                    print "Using country code"
                    break
            else:
                iter_timezone = model.get_value(iter, 1).name
                if iter_timezone == timezone:
                    combo.set_active_iter(iter)                        
                    print "Using system timezone name"
                    break
            iter = model.iter_next(iter)

